from trilogy_nlp.prompts.prompt_executor import FilterRefinementCase
from trilogy_nlp.models import FilterRefinementResponse
from tests.utility import generate_test_case, evaluate_cases
from trilogy.core.models import DataType


def gen_select_test(word: list[str]):
    def select_test(x: FilterRefinementResponse):
        print(x.new_values, word)
        return x.new_values == word

    return [select_test]


def test_filter_refinement(engine):
    case1 = generate_test_case(
        FilterRefinementCase,
        values=[
            "California",
        ],
        datatype=DataType.STRING,
        description="The two character abbreviation code for a state, such as MA for Massachusetts or CT for Connecticut.",
        tests=gen_select_test(
            [
                "CA",
            ],
        ),
        llm=engine,
    )

    case2 = generate_test_case(
        FilterRefinementCase,
        values=[
            "95%",
        ],
        datatype=DataType.FLOAT,
        description="Field storing a float representing the percentage of the population that likes coconuts, with a max of 1.0",
        tests=gen_select_test(
            [
                0.95,
            ],
        ),
        llm=engine,
    )
    evaluate_cases([case1, case2])



from trilogy_nlp.tokenization import tokens_to_concept


def test_tokens_to_concept():
    x = tokens_to_concept(
        ["date", "year", "time"],
        ["answer.creation.date", "answer.creation.date.year"],
        universe=["answer"],
        limits=1,
    )
    assert x == ["answer.creation.date.year"]

    x = tokens_to_concept(
        ["timestamp", "second", "time", "hour", "date", "minute"],
        [
            "badge.awarded_time",
            "badge.awarded_date",
            "answer.creation_date",
            "question.creation_date",
        ],
        universe=[
            "month",
            "timestamp",
            "second",
            "time",
            "hour",
            "count",
            "date",
            "minute",
            "year",
            "question",
        ],
        limits=1,
    )
    assert x == ["question.creation_date"]



from trilogy_nlp.prompts.prompt_executor import SemanticExtractionPromptCase
from trilogy_nlp.models import InitialParseResponse, FilterResult, OrderResult
from tests.utility import generate_test_case, evaluate_cases
from trilogy.core.enums import Ordering, ComparisonOperator
from pydantic import BaseModel

from typing import List


def flatten_arg_list(obj, args):
    output = []
    for arg in args:
        output += getattr(obj, arg)
    return output


def coerce_list(item):
    if not isinstance(item, list):
        return [item]
    return item


def validator_factory(key, test_values):
    def validator(input: InitialParseResponse, lkey=key, test_values=test_values):
        lkey = key
        ltest = coerce_list(test_values)
        field_values: list[str] = [x for x in coerce_list(getattr(input, lkey))]
        # if all of the values appear in at least one of the sub phrases
        check = all(
            [
                any(
                    (test == partial or (isinstance(partial, str) and test in partial))
                    for partial in field_values
                )
                for test in ltest
            ]
        )
        if not check:
            raise ValueError(
                "could not find ",
                test_values,
                " in ",
                key,
                " with ",
                field_values,
                "from",
                str(input),
            )
        return check

    return validator


def gen_validate_initial_parse_result(**kwargs):
    outputs = []
    for key, test_values in kwargs.items():
        # CRITICAL
        # we need to assign the values here in the loop to defaults on the lambda
        # to avoid lazy binding
        outputs.append(
            lambda x, test_values=test_values, key=key: validator_factory(
                key, test_values
            )(x)
        )
    return outputs


def test_extraction_prompt(test_logger, engine):
    case1 = generate_test_case(
        SemanticExtractionPromptCase,
        llm=engine,
        question="How many questions are asked per year? Order results by year desc",
        tests=gen_validate_initial_parse_result(
            selection=["question", "year", "count"],
            order=[OrderResult(concept="year", order=Ordering.DESCENDING)],
        ),
    )

    case2 = generate_test_case(
        SemanticExtractionPromptCase,
        question="How many questions were asked in the year 2020?",
        llm=engine,
        tests=gen_validate_initial_parse_result(
            selection=["question"],
            filtering=[
                FilterResult(
                    concept="year", values=["2020"], operator=ComparisonOperator.EQ
                )
            ],
        ),
    )

    case3 = generate_test_case(
        SemanticExtractionPromptCase,
        question="50 most common names by count in the state of VT in the year 2010?",
        llm=engine,
        tests=gen_validate_initial_parse_result(
            selection=[
                "name",
                "count",
                "state",
            ],
            limit=50,
            filtering=[
                FilterResult(concept="year", values=["2010"], operator="="),
                FilterResult(concept="state", values=["VT"], operator="="),
            ],
        ),
    )

    case4 = generate_test_case(
        SemanticExtractionPromptCase,
        question="What were the 50 most common names by count in Vermont in the year 2010?",
        llm=engine,
        tests=gen_validate_initial_parse_result(
            selection=["name", "count", "state", "year"],
            limit=50,
            filtering=[
                FilterResult(
                    concept="year", values=["2010"], operator=ComparisonOperator.EQ
                ),
                FilterResult(
                    concept="state", values=["Vermont"], operator=ComparisonOperator.EQ
                ),
            ],
        ),
    )
    evaluate_cases([case1, case2, case3, case4])


def test_like_predicates(engine):
    case1 = generate_test_case(
        SemanticExtractionPromptCase,
        question="SO answers where the body contains the word jaguar?",
        tests=gen_validate_initial_parse_result(
            selection=["body"],
            filtering=[
                MultiModeFilterMatch(
                    valid=[
                        FilterResult(
                            concept="body",
                            values=["%jaguar%"],
                            operator=ComparisonOperator.LIKE,
                        ),
                        FilterResult(
                            concept="post body",
                            values=["%jaguar%"],
                            operator=ComparisonOperator.LIKE,
                        ),
                    ]
                )
            ],
        ),
        llm=engine,
    )
    evaluate_cases([case1])


class MultiModeFilterMatch(BaseModel):
    valid: List[FilterResult]

    def __eq__(self, x):
        return any([x == v for v in self.valid])


def test_abstract_terms(engine):
    case1 = generate_test_case(
        SemanticExtractionPromptCase,
        question="Shoe sales on christmas day?",
        llm=engine,
        tests=gen_validate_initial_parse_result(
            selection=["product", "sale"],
            filtering=[
                MultiModeFilterMatch(
                    valid=[
                        FilterResult(
                            concept="product",
                            values=["Shoe"],
                            operator=ComparisonOperator.EQ,
                        ),
                        FilterResult(
                            concept="day",
                            values=["Christmas"],
                            operator=ComparisonOperator.EQ,
                        ),
                        FilterResult(
                            concept="day",
                            values=["Christmas Day"],
                            operator=ComparisonOperator.EQ,
                        ),
                        FilterResult(
                            concept="day of week",
                            values=["Christmas Day"],
                            operator=ComparisonOperator.EQ,
                        ),
                        FilterResult(
                            concept="day of year",
                            values=["December 25"],
                            operator=ComparisonOperator.EQ,
                        ),
                        FilterResult(
                            concept="day of year",
                            values=["12/25"],
                            operator=ComparisonOperator.EQ,
                        ),
                    ]
                )
            ],
        ),
    )
    evaluate_cases([case1])



import uuid
from typing import List, Union
from trilogy.core.enums import BooleanOperator, Purpose
from trilogy.core.query_processor import process_query
from trilogy.core.models import (
    Comparison,
    Concept,
    Conditional,
    DataType,
    Environment,
    OrderBy,
    Ordering,
    OrderItem,
    SelectStatement,
    WhereClause,
    ProcessedQuery,
    unique,
)

from trilogy_nlp.constants import DEFAULT_LIMIT, logger
from trilogy_nlp.models import (
    FilterResult,
    FinalFilterResult,
    FinalOrderResult,
    InitialParseResponse,
    IntermediateParseResults,
    OrderResult,
    # TokenInputs,
    SemanticTokenResponse,
    FinalParseResponse,
)
from trilogy_nlp.prompts import (
    FilterRefinementCase,
    FilterRefinementErrorCase,
    SelectionPromptCase,
    SemanticExtractionPromptCase,
    SemanticToTokensPromptCase,
    run_prompt,
)
from trilogy_nlp.tokenization import build_token_list_by_purpose, tokens_to_concept
from langchain_core.language_models import BaseLanguageModel
from uuid import UUID


def get_phrase_from_x(x: Union[str, FilterResult, OrderResult]):
    if isinstance(x, str):
        return x
    elif isinstance(x, FilterResult):
        if len(x.values) == 1:
            return f"{x.concept} {x.values[0]}"
        return f"{x.concept} {x.values}"
    elif isinstance(x, OrderResult):
        return x.concept
    raise ValueError


def tokenize_phrases(
    purpose: str,
    phrase_list: List[str],
    tokens: List[str],
    session_uuid,
    log_info: bool,
    llm: BaseLanguageModel,
) -> SemanticTokenResponse:
    phrase_tokens: SemanticTokenResponse = run_prompt(  # type: ignore
        SemanticToTokensPromptCase(
            phrases=phrase_list,
            tokens=tokens,
            purpose=purpose,
            llm=llm,
        ),
        debug=True,
        session_uuid=session_uuid,
        log_info=log_info,
    )

    return phrase_tokens


def concept_names_from_token_response(
    phrase_tokens: SemanticTokenResponse,
    concepts: dict[str, Concept],
    token_universe: list | None,
) -> list[str]:
    token_universe_internal = token_universe or []
    output: list[str] = []

    for mapping in phrase_tokens:
        token_universe_internal += mapping.tokens
    token_universe_internal = list(set(token_universe_internal))
    for mapping in phrase_tokens.root:
        found = False
        concepts_str_matches = tokens_to_concept(
            mapping.tokens,
            [c for c in concepts.keys()],
            limits=50,
            universe=token_universe_internal,
        )
        logger.debug(f"For phrase {mapping.phrase} got {concepts_str_matches}")
        if concepts_str_matches:
            # logger.info(f"For phrase {mapping.phrase} got {concepts_str_matches}")
            output += concepts_str_matches
            found = True
            continue
        if not found:
            raise ValueError(
                f"Could not find concept for input {mapping.phrase} with tokens {mapping.tokens} and concepts {list(concepts.keys())}"
            )
    return output


def coerce_values(input: List[Union[str, int, float, bool]], dtype=DataType):
    if dtype == DataType.INTEGER:
        return [int(x) for x in input]
    elif dtype == DataType.FLOAT:
        return [float(x) for x in input]
    elif dtype == DataType.BOOL:
        return [bool(x) for x in input]
    return input


def enrich_filter_parent(
    input: FinalFilterResult,
    log_info: bool,
    llm: BaseLanguageModel,
    session_uuid: UUID | None,
):
    try:
        enrich_filter(
            input=input,
            log_info=log_info,
            llm=llm,
            session_uuid=session_uuid,
        )
    except ValueError:
        return


def enrich_filter(
    input: FinalFilterResult,
    log_info: bool,
    llm: BaseLanguageModel,
    session_uuid: UUID | None = None,
):
    if not (input.concept.metadata and input.concept.metadata.description):
        # coerce even without description
        try:
            input.values = coerce_values(input.values, input.concept.datatype)
            return input
        except ValueError as e:
            input.values = coerce_values(
                run_prompt(  # type: ignore
                    FilterRefinementErrorCase(
                        values=input.values,
                        error=str(e),
                        datatype=input.concept.datatype,
                        llm=llm,
                    ),
                    session_uuid=session_uuid,
                    log_info=log_info,
                ).new_values,
                input.concept.datatype,
            )
            return input

    try:
        input.values = coerce_values(
            run_prompt(  # type: ignore
                FilterRefinementCase(
                    values=input.values,
                    description=input.concept.metadata.description,
                    datatype=input.concept.datatype,
                    llm=llm,
                ),
                session_uuid=session_uuid,
                log_info=log_info,
            ).new_values,
            input.concept.datatype,
        )
    except ValueError as e:
        input.values = coerce_values(
            run_prompt(  # type: ignore
                FilterRefinementErrorCase(
                    values=input.values,
                    error=str(e),
                    datatype=input.concept.datatype,
                    llm=llm,
                ),
                session_uuid=session_uuid,
                log_info=log_info,
            ).new_values,
            input.concept.datatype,
        )
        return input


def discover_inputs(
    input_text: str,
    llm: BaseLanguageModel,
    input_environment: Environment,
    debug: bool = False,
    log_info: bool = True,
) -> IntermediateParseResults:
    # the core logic flow

    # DETERMINSTIC: setup logging
    # LLM: semantic extraction
    # LLM: tokenization of all strings (reduce search space over all concepts)
    # DETERMINISTIC: concept candidates from tokens (map reduced search space to candidates)
    # LLM: final selection of concepts and mapping to output roles
    # LLM: enrich filter values
    # DETERMINISTIC: return results

    # DETERMINISTIC: setup logging
    session_uuid = uuid.uuid4()
    env_concepts = input_environment.concepts

    # LLM: semantic extraction
    parsed: InitialParseResponse = run_prompt(  # type:ignore
        SemanticExtractionPromptCase(input_text, llm=llm),
        debug=debug,
        log_info=log_info,
        session_uuid=session_uuid,
    )

    # LLM: tokenization of all strings (reduce search space over all concepts)
    concept_candidates = []
    token_response_mapping: dict[str, SemanticTokenResponse] = {}
    for semantic_category, valid_purposes in {
        "metrics": [Purpose.METRIC],
        "dimensions": [Purpose.KEY, Purpose.PROPERTY, Purpose.CONSTANT],
        "filtering": list(Purpose),
        "order": list(Purpose),
    }.items():
        category_tokens = build_token_list_by_purpose(env_concepts, valid_purposes)
        local_phrases = [
            get_phrase_from_x(x) for x in getattr(parsed, semantic_category)
        ]
        # skip if we have no relevant phrases
        if not local_phrases:
            continue
        if not category_tokens:
            raise ValueError("No tokens for category")
        token_mapping = tokenize_phrases(
            semantic_category,
            local_phrases,
            category_tokens,
            log_info=log_info,
            session_uuid=session_uuid,
            llm=llm,
        )
        token_response_mapping[semantic_category] = token_mapping

    # DETERMINISTIC - generate concept candidates from tokens (map reduced search space to candidates)
    token_universe = set()
    for _, token_mapping_pass_one in token_response_mapping.items():
        for mapping in token_mapping_pass_one:
            for t in mapping.tokens:
                token_universe.add(t)
    token_universe_list = list(token_universe)

    for _, token_mapping_pass_two in token_response_mapping.items():
        for item in token_mapping_pass_two:
            if not all([x in category_tokens] for x in item.tokens):
                invalid = [x for x in item.tokens if x not in category_tokens]
                raise ValueError(
                    f"Phrase {item.phrase} return invalid tokens {invalid}"
                )
        concept_candidates += concept_names_from_token_response(
            token_mapping_pass_two, env_concepts, token_universe=token_universe_list
        )

    # LLM - use our concept candidates to generate the final output
    selections: FinalParseResponse = run_prompt(  # type: ignore
        SelectionPromptCase(
            concept_names=concept_candidates,
            all_concept_names=list(env_concepts.keys()),
            question=input_text,
            llm=llm,
        ),
        debug=debug,
        session_uuid=session_uuid,
        log_info=log_info,
    )
    selected_concepts = list(set(selections.selection))

    final_ordering = [
        FinalOrderResult(concept=env_concepts[order.concept], order=order.order)
        for order in selections.order
    ]

    final_filters_pre = [
        FinalFilterResult(
            concept=env_concepts[filter.concept],
            operator=filter.operator,
            values=filter.values,
        )
        for filter in selections.filtering
    ]

    # LLM: enrich filter values
    for item in final_filters_pre:
        enrich_filter(item, log_info=log_info, session_uuid=session_uuid, llm=llm)

    # DETERMINISTIC: return results
    return IntermediateParseResults(
        select=[env_concepts[c] for c in selected_concepts],
        limit=parsed.limit or 20,
        order=final_ordering,
        filtering=final_filters_pre,
    )


def safe_limit(input: int | None) -> int:
    if not input:
        return DEFAULT_LIMIT
    if input in (-1, 0):
        return DEFAULT_LIMIT
    return input


def parse_order(
    input_concepts: List[Concept], ordering: List[FinalOrderResult] | None
) -> OrderBy:
    default_order = [
        OrderItem(expr=c, order=Ordering.DESCENDING)
        for c in input_concepts
        if c.purpose == Purpose.METRIC
    ]
    if not ordering:
        return OrderBy(items=default_order)
    final = []
    for order in ordering:
        final.append(OrderItem(expr=order.concept, order=order.order))
    return OrderBy(items=final)


def parse_filter(input: FinalFilterResult) -> Comparison | None:
    return Comparison(
        left=input.concept,
        right=input.values[0] if len(input.values) == 1 else input.values,
        operator=input.operator,
    )


def parse_filtering(filtering: List[FinalFilterResult]) -> WhereClause | None:
    base = []
    for item in filtering:
        parsed = parse_filter(item)
        if parsed:
            base.append(parsed)
    if not base:
        return None
    if len(base) == 1:
        return WhereClause(conditional=base[0])
    left: Conditional | Comparison = base.pop()
    while base:
        right = base.pop()
        new = Conditional(left=left, right=right, operator=BooleanOperator.AND)
        left = new
    return WhereClause(conditional=left)


def parse_query(
    input_text: str,
    llm: BaseLanguageModel,
    input_environment: Environment,
    debug: bool = False,
    log_info: bool = True,
):
    intermediate_results = discover_inputs(
        input_text=input_text,
        input_environment=input_environment,
        debug=debug,
        log_info=log_info,
        llm=llm,
    )
    selection = unique(intermediate_results.select, "address")

    order = parse_order(selection, intermediate_results.order)

    filtering = parse_filtering(intermediate_results.filtering)
    # from trilogy.core.models import unique
    # concepts = unique(concepts, 'address')
    if debug:
        print("Concepts found")
        for c in intermediate_results.select:
            print(c.address)
    query = SelectStatement(
        selection=selection,
        limit=safe_limit(intermediate_results.limit),
        order_by=order,
        where_clause=filtering,
    )
    return query


def build_query(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    log_info: bool = True,
) -> ProcessedQuery:
    query = parse_query(
        input_text=input_text,
        input_environment=input_environment,
        llm=llm,
        debug=debug,
        log_info=log_info,
    )
    return process_query(statement=query, environment=input_environment)



from .monkeypatch import patch_promptimize, patch_langchain

patch_promptimize()
patch_langchain()

from .main import build_query  # noqa: E402
from .enums import Provider  # noqa: E402
from .core import NLPEngine  # noqa: E402

__version__ = "0.0.18"

__all__ = ["build_query", "Provider", "NLPEngine"]



from typing import Iterable, Dict, List
from trilogy.core.models import Concept
from trilogy.core.enums import Purpose
import re
from collections import defaultdict


class Token(str):
    def __init__(*args):
        super.__init__(*args)


def split_to_tokens(input_text: str) -> list[str]:
    return list(set(re.split("\.|\_", input_text)))


def build_token_list_by_purpose(
    concepts: Dict[str, Concept], purposes: Iterable[Purpose]
) -> list[str]:
    concepts = {k: v for k, v in concepts.items() if v.purpose in purposes}

    unique = set(concepts.keys())
    final = set()
    for concept in unique:
        # matched = [comparison  for comparison in unique if comparison in concept]
        # filtered = [m for m in matched if m !=concept]
        # if not filtered:
        #     final.append(concept)
        for x in split_to_tokens(concept):
            final.add(x)
    return list(final)


def tokens_to_concept(
    tokens: list[str],
    concepts: List[str],
    limits: int = 5,
    universe: list[str] | None = None,
) -> list[str] | None:
    tokens = list(set(tokens))
    universe_list = list(set(universe or []))
    mappings = {x: split_to_tokens(x) for x in concepts}
    candidates = [x for x in concepts if any(token in mappings[x] for token in tokens)]
    pickings = defaultdict(list)

    if not candidates:
        return None
    tiers = set()

    for candidate in candidates:
        score = 0
        candidate_tokens = mappings[candidate]
        # exact match to concept is +2
        for x in tokens:
            if x in candidate_tokens:
                # print(f'+2 score from token {x} in candidate list')
                score += 10
        # universe context is + 1
        for y in universe_list:
            if y in candidate_tokens:
                # print(f"+1 score from token {y} in universe")
                score += 1
        pickings[score].append(candidate)
        tiers.add(score)
    tier_list = sorted(list(tiers), key=lambda x: -x)
    found: List[str] = []
    while len(found) < limits and tier_list:
        stier = tier_list.pop(0)
        # sort within list by length ascending, then alphabetical
        # for consistency
        candidates = sorted(pickings[stier], key=lambda x: (len(x), x))
        required = limits - len(found)
        found += candidates[:required]
    return found



# Brining some "prompt generator" classes - note that you can derive and extend those
from promptimize.prompt_cases import TemplatedPromptCase

# Bringing some useful eval function that help evaluating and scoring responses
# eval functions have a handle on the prompt object and are expected
# to return a score between 0 and 1
from trilogy_nlp.constants import logger
from trilogy_nlp.models import (
    InitialParseResponse,
    ConceptSelectionResponse,
    SemanticTokenResponse,
    FilterRefinementResponse,
    FinalParseResponse,
)
from trilogy.core.models import DataType, ListType, MapType, StructType
from trilogy_nlp.cache_providers.base import BaseCache
from trilogy_nlp.cache_providers.local_sqlite import SqlliteCache
from pydantic import BaseModel, ValidationError
from typing import List, Optional, Callable, Union, Type, overload
import uuid
import json
import os
from jinja2 import FileSystemLoader, Environment, Template
from os.path import dirname
from langchain_core.messages import HumanMessage
from trilogy_nlp.helpers import retry_with_exponential_backoff
from langchain_core.language_models import BaseLanguageModel
import sys
from pathlib import Path

PROMPT_STOPWORD = "<EOM>"

MAX_REFINEMENT_TRIES = 4


def resource_path() -> str:
    if not getattr(sys, "frozen", False):
        return dirname(__file__)
    # If the application is run as a bundle, the PyInstaller bootloader
    # extends the sys module by a flag frozen=True and sets the app
    # path into variable _MEIPASS'.
    application_path = Path(sys._MEIPASS)  # type: ignore
    search_path = application_path / "preql_nlp" / "prompts"

    return str(Path(search_path))


loader = FileSystemLoader(searchpath=resource_path())
templates = Environment(loader=loader)


def gen_hash(obj, keys: set[str]) -> str:
    """Generate a deterministic hash for an object across multiple runs"""
    import hashlib

    m = hashlib.sha256(usedforsecurity=False)

    # we need to hash things in a deterministic order
    key_list = sorted(list(keys), key=lambda x: x)
    for key in key_list:
        x = getattr(obj, key)
        if isinstance(x, Template):
            s = x.render()
        else:
            s = str(x)

        m.update(s.encode("utf-8"))

    return m.digest().hex()


class BasePreqlPromptCase(TemplatedPromptCase):
    parse_model: Type[BaseModel]
    template: Template
    stopword = PROMPT_STOPWORD

    def __init__(
        self,
        category: str,
        llm: BaseLanguageModel,
        fail_on_parse_error: bool = True,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        # this isn't actually the right way to pass through a stopword to the complete prompt
        # so we're splitting in the response
        # TODO: figure out how to do this when we figure out how to group prompts in one API call
        self.llm = llm
        super().__init__(
            category=category,
            evaluators=evaluators,
            prompt_executor_kwargs={"stopword": PROMPT_STOPWORD},
        )
        self._prompt_hash = str(uuid.uuid4())
        self.parsed = None
        self.fail_on_parse_error = fail_on_parse_error
        self.retry_prompt = "User: That response was incorrectly formatted. Please return the same content as the first response with formatting fixed (eg. valid JSON) to conform with original request. System:\n"
        self.stash: BaseCache = SqlliteCache()
        self.has_rerun = False

    @classmethod
    def parse_response(cls, response: str):
        try:
            return cls.parse_model.model_validate_json(response.split(cls.stopword)[0])
        except ValidationError as e:
            raise e

    def get_prompt_executor(self):
        if self.llm:
            return self.llm
        from langchain_openai import ChatOpenAI

        model_name = os.environ.get("OPENAI_MODEL") or "gpt-3.5-turbo"
        openai_api_key = os.environ.get("OPENAI_API_KEY")
        return ChatOpenAI(model=model_name, openai_api_key=openai_api_key)

    @retry_with_exponential_backoff
    def execute_prompt(self, prompt_str: str, skip_cache: bool = False):
        # if we already have a local result
        # skip hitting remote
        # TODO: make the cache provider pluggable and injected
        hash_val = gen_hash(self, self.attributes_used_for_hash)
        if not skip_cache:
            resp = self.stash.retrieve(hash_val)
            if resp:
                self.response = resp
                return self.response
        self.response = self.prompt_executor(
            [
                HumanMessage(content=prompt_str),
            ]
        ).content

        return self.response

    def get_extra_template_context(self):
        return {"stopword": self.stopword}

    def rerun(self):
        self.prompt = self.prompt + self.response + "\n" + self.retry_prompt
        self.has_rerun = True
        self.execute_prompt(self.prompt, skip_cache=True)

    def post_run(self):
        try:
            self.parsed = self.parse_response(self.response)
            logger.info(
                f"Sucessfully parsed response to {self.category}: {self.parsed.model_dump_json()}"
            )
            hash_val = gen_hash(self, self.attributes_used_for_hash)
            self.stash.store(hash_val, self.category, self.response)
        except ValidationError as e:
            if not self.has_rerun:
                self.rerun()
                return self.post_run()
            print(self.render())
            print("was unable to parse response using ", str(self.parse_model))
            print(self.response)
            if self.fail_on_parse_error:
                raise e

    def render(
        self,
    ):
        return self.template.render(**self.jinja_context)


class SemanticExtractionPromptCase(BasePreqlPromptCase):
    template = templates.get_template("prompt_query_semantic_groupings.jinja2")
    parse_model = InitialParseResponse

    attributes_used_for_hash = {
        "category",
        "question",
        "template",
    }

    def __init__(
        self,
        question: str,
        llm: BaseLanguageModel,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        self.question = question
        super().__init__(category="semantic_extraction", evaluators=evaluators, llm=llm)

    def get_extra_template_context(self):
        return {**super().get_extra_template_context(), "question": self.question}


class SemanticToTokensPromptCase(BasePreqlPromptCase):
    template = templates.get_template("prompt_semantic_to_tokens.jinja2")
    parse_model = SemanticTokenResponse

    attributes_used_for_hash = {
        "tokens",
        "phrases",
        "category",
        "template",
        "purpose",
    }

    def __init__(
        self,
        purpose: str,
        tokens: List[str],
        phrases: List[str],
        llm: BaseLanguageModel,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        tokens = [token for token in tokens if token]
        self.tokens = sorted(tokens)
        self.purpose = purpose
        # we need to ensure that we always have a list
        # for some LLMs to understand the prompt properly
        # pad out the inputs with a random phrase
        self.padding_phrase = "democratic elections"
        self.phrases = sorted(phrases + [self.padding_phrase])
        super().__init__(category="semantic_to_tokens", llm=llm, evaluators=evaluators)
        self.retries = 0

    def get_extra_template_context(self):
        return {
            **super().get_extra_template_context(),
            "purpose": self.purpose,
            "tokens": ", ".join([f'"{c}"' for c in self.tokens if c]),
            "phrase_str": ", ".join([f'"{c}"' for c in self.phrases]),
        }

    def _process_tokens(self, input_tokens):
        input_tokens.tokens = [t for t in input_tokens.tokens if t in self.tokens]
        return input_tokens

    def post_run(self):
        super().post_run()
        self.parsed.root = [
            self._process_tokens(x)
            for x in self.parsed.root
            if x.phrase != self.padding_phrase
        ]
        # TODO: evaluate tradeoffs
        # do we ask to refine, or just silently drop invalid tokens?
        # tokens = []
        # for x in self.parsed.__root__:
        #     tokens+= x.tokens
        # missing = False
        # tokens = list(set(tokens))
        # for token in tokens:
        #     if token not in self.tokens:
        #         missing = True
        #         break
        # if missing:
        #     self.retries += 1
        #     valid = ", ".join([f'"{c}"' for c in self.tokens if c]),
        #     retry_prompt = f'User: The token "{token}" in your answer was not in the provided list, please return a new answer without any unprovided lists. Valid tokens are [{valid}] Return only the corrected JSON, do not apologize. \nSystem: '
        #     self.prompt = self.prompt + "\n" + self.response + "\n" + retry_prompt
        #     if self.retries < MAX_REFINEMENT_TRIES:
        #         self.execute_prompt(self.prompt, skip_cache=True)
        #         return self.post_run()
        #     else:
        #         print(self.prompt)
        #         raise ValueError(
        #             f"After {self.retries} attempts, LLM returned token {token} that does not exist in input tokens, cannot progress - returned {self.parsed}"
        #         )


class SelectionPromptCase(BasePreqlPromptCase):
    template = templates.get_template("prompt_final_concepts_v2.jinja2")
    parse_model = FinalParseResponse

    attributes_used_for_hash = {
        "question",
        "concept_names",
        "category",
        "template",
    }

    def __init__(
        self,
        question: str,
        concept_names: List[str],
        llm: BaseLanguageModel,
        all_concept_names: List[str] | None = None,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        self.question = question
        self.all_concept_names_internal = all_concept_names or concept_names
        self.concept_names = sorted(list(set(concept_names)), key=lambda x: x)
        super().__init__(
            evaluators=evaluators, llm=llm, category="final_concept_selection"
        )
        self.execution.score = None
        self.retries = 0

    def get_extra_template_context(self):
        return {
            **super().get_extra_template_context(),
            "concept_string": ", ".join([f'"{c}"' for c in self.concept_names]),
            "question": self.question,
        }

    def post_run(self):
        super().post_run()

        selected_concepts: List[str] = self.parsed.selection
        for item in self.parsed.filtering:
            selected_concepts.append(item.concept)
        for item in self.parsed.order:
            selected_concepts.append(item.concept)

        missing = False
        selected_concepts = list(set(selected_concepts))
        for selection in selected_concepts:
            if selection not in self.all_concept_names_internal:
                missing = True
                break
        if missing:
            self.retries += 1
            valid = ", ".join([f'"{c}"' for c in self.concept_names])
            retry_prompt = f'User: The concept "{selection}" in your answer was not one in the provided valid list, please return a new answer with only valid concepts from this list: [{valid}] Return only the corrected JSON, do not apologize. \nSystem: '
            self.prompt = self.prompt + "\n" + self.response + "\n" + retry_prompt
            if self.retries < MAX_REFINEMENT_TRIES:
                self.execute_prompt(self.prompt, skip_cache=True)
                return self.post_run()
            else:
                raise ValueError(
                    f"LLM returned concept {selection} that does not exist in input concepts, cannot progress - returned {self.parsed}"
                )


class FilterRefinementCase(BasePreqlPromptCase):
    template = templates.get_template("prompt_refine_filter.jinja2")
    parse_model = FilterRefinementResponse

    attributes_used_for_hash = {
        "values",
        "description",
        "template",
    }

    def __init__(
        self,
        values: list[str | int | float | bool],
        description: str,
        datatype: DataType | StructType | ListType | MapType,
        llm: BaseLanguageModel,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        self.values = values
        self.description = description
        self.datatype = datatype
        super().__init__(evaluators=evaluators, category="filter_refinement", llm=llm)

    def get_extra_template_context(self):
        return {
            **super().get_extra_template_context(),
            "values": ", ".join([f'"{x}"' for x in self.values]),
            "description": self.description,
            "datatype": self.datatype.value,
        }


class FilterRefinementErrorCase(BasePreqlPromptCase):
    template = templates.get_template("prompt_refine_filter_error.jinja2")
    parse_model = FilterRefinementResponse

    attributes_used_for_hash = {
        "values",
        "error",
        "template",
    }

    def __init__(
        self,
        values: list[str | int | float | bool],
        error: str,
        datatype: DataType | StructType | ListType | MapType,
        llm: BaseLanguageModel,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
    ):
        self.values = values
        self.error = error
        self.datatype = datatype
        super().__init__(
            evaluators=evaluators, category="filter_refinement_error", llm=llm
        )

    def get_extra_template_context(self):
        return {
            **super().get_extra_template_context(),
            "values": ", ".join([f'"{x}"' for x in self.values]),
            "error": self.error,
            "datatype": self.datatype.value,
        }


DATA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "log_data"
)
if not os.path.exists(DATA_DIR):
    os.makedirs(DATA_DIR)


def log_prompt_info(prompt: TemplatedPromptCase, session_uuid: uuid.UUID):
    prompt_hash = prompt.prompt_hash
    prompt_context = prompt.jinja_context
    template = prompt.template.render(**prompt.jinja_context)
    category = prompt.category

    data = {
        "prompt_token": prompt_hash,
        "prompt_context": prompt_context,
        "template": template,
        "category": category,
        "session_uuid": str(session_uuid),
        "response": prompt.response,
    }
    path = os.path.join(
        DATA_DIR, str(session_uuid), category + "-" + prompt_hash + ".json"
    )
    with open(
        path,
        "w",
    ) as f:
        logger.debug("printing to...{}".format(path))
        json.dump(data, f)


@overload
def run_prompt(  # type: ignore
    prompt: SelectionPromptCase,
    debug: bool = False,
    log_info: bool = True,
    session_uuid: uuid.UUID | None = None,
) -> ConceptSelectionResponse:
    ...


@overload
def run_prompt(  # type: ignore
    prompt: SemanticExtractionPromptCase,
    debug: bool = False,
    log_info: bool = True,
    session_uuid: uuid.UUID | None = None,
) -> InitialParseResponse:
    ...


@overload
def run_prompt(  # type: ignore
    prompt: SemanticToTokensPromptCase,
    debug: bool = False,
    log_info: bool = True,
    session_uuid: uuid.UUID | None = None,
) -> SemanticTokenResponse:
    ...


@overload
def run_prompt(  # type: ignore
    prompt: FilterRefinementCase,
    debug: bool = False,
    log_info: bool = True,
    session_uuid: uuid.UUID | None = None,
) -> FilterRefinementResponse:
    ...


def run_prompt(
    prompt: BasePreqlPromptCase,
    debug: bool = False,
    log_info: bool = True,
    session_uuid: uuid.UUID | None = None,
) -> (
    ConceptSelectionResponse
    | SemanticTokenResponse
    | InitialParseResponse
    | FilterRefinementResponse
):
    if not session_uuid:
        session_uuid = uuid.uuid4()

    if log_info and not os.path.exists(os.path.join(DATA_DIR, str(session_uuid))):
        os.makedirs(os.path.join(DATA_DIR, str(session_uuid)))
    if debug:
        logger.debug("prompt")
        logger.debug(prompt.__dict__)
        logger.debug(prompt.render())

    raw_output = prompt._run(dry_run=False)
    if debug:
        logger.debug("output")
        logger.debug(raw_output)

    if log_info:
        log_prompt_info(prompt, session_uuid)
    if prompt.parsed:
        return prompt.parsed
    raise ValueError("Could not parse prompt response!")



from typing import Optional
from langchain.agents import create_structured_chat_agent, AgentExecutor
from trilogy_nlp.main import safe_limit
from trilogy.core.models import (
    Concept,
    Environment,
    ProcessedQuery,
    SelectStatement,
    Comparison,
    Conditional,
    OrderBy,
    OrderItem,
    WhereClause,
)
from typing import List
from trilogy.core.query_processor import process_query
from langchain.tools import Tool
from datetime import datetime
import json
from pydantic import BaseModel, field_validator
from trilogy.core.enums import ComparisonOperator, Ordering, Purpose, BooleanOperator
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.language_models import BaseLanguageModel
from langchain_community.tools.wikidata.tool import WikidataAPIWrapper, WikidataQueryRun

langchain_chat_kwargs = {
    "temperature": 0,
    "max_tokens": 4000,
    "verbose": True,
}
chat_openai_model_kwargs = {
    "top_p": 1.0,
    "frequency_penalty": 0.0,
    "presence_penalty": -1,
}


class FilterResultV2(BaseModel):
    """The result of the filter prompt"""

    column: str
    values: list[str | int | float | bool]
    operator: ComparisonOperator

    @field_validator("values", mode="plain")
    @classmethod
    def values_validation(cls, v):
        if isinstance(v, list):
            return v
        return [v]


class OrderResultV2(BaseModel):
    """The result of the order prompt"""

    column: str
    order: Ordering


class InitialParseResponseV2(BaseModel):
    """The result of the initial parse"""

    columns: list[str]
    limit: Optional[int] = 100
    order: Optional[list[OrderResultV2]] = None
    filtering: Optional[list[FilterResultV2]] = None

    @property
    def selection(self) -> list[str]:
        filtering = [f.column for f in self.filtering] if self.filtering else []
        order = [x.column for x in self.order] if self.order else []
        return list(set(self.columns + filtering + order))

    @field_validator("filtering", mode="plain")
    @classmethod
    def filtering_validation(cls, v):
        if isinstance(v, dict):
            return [FilterResultV2.model_validate(v)]
        return [FilterResultV2.model_validate(x) for x in v]

    @field_validator("order", mode="plain")
    @classmethod
    def order_validation(cls, v):
        if isinstance(v, dict):
            return [OrderResultV2.model_validate(v)]
        return [OrderResultV2.model_validate(x) for x in v]


def parse_order(
    input_concepts: List[Concept], ordering: List[OrderResultV2] | None
) -> OrderBy:
    default_order = [
        OrderItem(expr=c, order=Ordering.DESCENDING)
        for c in input_concepts
        if c.purpose == Purpose.METRIC
    ]
    if not ordering:
        return OrderBy(items=default_order)
    final = []
    for order in ordering:
        concept = [x for x in input_concepts if x.address == order.column][0]
        final.append(OrderItem(expr=concept, order=order.order))
    return OrderBy(items=final)


def parse_filter(
    input_concepts: List[Concept], input: FilterResultV2
) -> Comparison | None:
    concept = [x for x in input_concepts if x.address == input.column][0]
    return Comparison(
        left=concept,
        right=input.values[0] if len(input.values) == 1 else input.values,
        operator=input.operator,
    )


def parse_filtering(
    input_concepts: List[Concept], filtering: List[FilterResultV2]
) -> WhereClause | None:
    base = []
    for item in filtering:
        parsed = parse_filter(input_concepts, item)
        if parsed:
            base.append(parsed)
    if not base:
        return None
    if len(base) == 1:
        return WhereClause(conditional=base[0])
    left: Conditional | Comparison = base.pop()
    while base:
        right = base.pop()
        new = Conditional(left=left, right=right, operator=BooleanOperator.AND)
        left = new
    return WhereClause(conditional=left)


def run_query_save_results(executor, query: str):
    executor.execute_query(query)


def validate_query(query: str, environment: Environment):
    parsed = InitialParseResponseV2.model_validate_json(query)
    for x in parsed.columns:
        if x not in environment.concepts:
            return {"status": "invalid", "error": f"{x} in fields not in concepts"}
    if parsed.order:
        for y in parsed.order:
            if y.column not in environment.concepts:
                return {
                    "status": "invalid",
                    "error": f"{y.column} in order not in concepts",
                }
    if parsed.filtering:
        for z in parsed.filtering:
            if z.column not in environment.concepts:
                return {
                    "status": "invalid",
                    "error": f"{z.column} in filtering not in concepts",
                }
    return {"status": "valid"}


def get_model_description(query: str, environment: Environment):
    """
    Get the description of the dataset.
    """
    datasources = ", ".join(
        set([x.identifier for x in environment.datasources.values()])
    )
    return json.dumps(
        {
            "description": f"database contains information about: {datasources}. No more specific information is available. Use the fields tool to get more specific information."
        }
    )


def get_today_date(query: str) -> str:
    """
    Useful to get the date of today.
    """
    # Getting today's date in string format
    today_date_string = datetime.now().strftime("%Y-%m-%d")
    return today_date_string


def concept_to_string(concept: Concept):
    return concept.address


def get_fields(environment: Environment, search: str, *args, **kwargs) -> str:
    return json.dumps(
        {
            "fields": [
                concept_to_string(x)
                for x in environment.concepts.values()
                if search.lower() in x.address.lower()
            ]
        }
    )


def sql_agent_tools(environment):
    tools = [
        Tool.from_function(
            func=lambda x: get_model_description(x, environment),
            name="get_database_description",
            description="""
           Describe the database and general groupings of fields available.""",
        ),
        Tool.from_function(
            func=lambda x: validate_query(x, environment),
            name="validate_response",
            description="""
            Check that your response is formatted properly and accurate before your final answer.
            """,
        ),
        Tool.from_function(
            func=lambda x: get_fields(environment, x),
            name="get_fields",
            description="""
            The list of fields that can be selected that contain the input string. When looking for an aggregate (count, etc) prefer concepts with that string in the name.
            """,
        ),
        Tool.from_function(
            func=get_today_date,
            name="get_today_date",
            description="""
            Useful to get the date of today.
            """,
        ),
    ]
    return tools


def parse_query(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    log_info: bool = True,
):
    system = """Thought Process: You are a data analyst asstant. Your job is to get questions from 
    the analyst and tell them how to write a 
    SQL query to answer them in a step by step fashion. 

    You can get information on the columns available and cannot create any new ones;
    do not worry about tables, the analyst will join them.

    If you need additional contextual information that would not be in the database, use the wikimedia tool to get it.

    Do your best to get to the most complete answer possible using all tools. 

    Your goal will be to create a summary of steps in JSON format for your analyst.
    The final information should be a VALID JSON blob with the following keys and values followed by a stopword: <EOD>:
    - columns: a list of columns
    - limit: a number of records to limit the results to, -1 if none specified
    - order: a list of columns to order the results by, with the option to specify ascending or descending
    -- column: a column name to order by
    -- order: the direction of ordering, "asc" or "desc"
    - filtering: a list of all objects to filter the results on, where each object has the following keys:
    -- column: a column to filter on
    -- values: the value the column is filtered to
    -- operator: the comparison operator, one of "=", "in", "<", ">", "<=", "like", or ">=". A range, or between, should be expressed as two inequalities. 


    You should always call the the validate_response tool on what you think is the final answer before returning the "Final Answer" action.
    You have access to the following tools:

    {tools}

    Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input). 
    You will get valuable information from using tools before producing your final answer.

    Use as many tools as needed before producing the "Final Answer" action.

    Valid "action" values: any of {tool_names} and, for your last result "Final Answer". 
    Only return "Final Answer" when you are done with all work. Nnever set the action to 
    "Final Answer" before you are done. 

    You should always call the the validate_response tool with your final answer before sending it to the CEO.

    Provide only ONE action per $JSON_BLOB, as shown:

    ```
    {{
        "action": $TOOL_NAME,
        "action_input": $INPUT
    }}
    ```

    Follow this format in responses:

    Question: input question to answer
    Thought: consider previous and subsequent steps
    Action:
    ```
    $JSON_BLOB
    ```
    Observation: action result
    ... (repeat Thought/Action/Observation N times)
    Thought: I know what to respond
    Action:
    ```
    {{
        "action": "get_database_description",
        "action_input": "some_dataset"
    }}

    Once you have used any tools (listed below) as needed, you will produce your final result in this format. After producing your
    final answer, you cannot take any more steps.

    {{
        "action": "Final Answer",
        "action_input": <VALID_JSON_SPEC_DEFINED_ABOVE>
    }}

    Begin! Reminder to ALWAYS respond with a valid json blob of an action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation"""

    human = """{input}

    {agent_scratchpad}

    (reminder to respond in a JSON blob no matter what)"""

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", system),
            MessagesPlaceholder("chat_history", optional=True),
            ("human", human),
        ]
    )
    wiki = WikidataQueryRun(api_wrapper=WikidataAPIWrapper())  # type: ignore
    wiki.description = (
        "Look up information on a specific string from Wikipedia. Use to get context"
    )
    tools = sql_agent_tools(input_environment) + [wiki]
    chat_agent = create_structured_chat_agent(
        llm=llm,
        tools=tools,
        prompt=prompt,
    )
    agent_executor = AgentExecutor(
        agent=chat_agent, tools=tools, verbose=True, handle_parsing_errors=True  # type: ignore
    )

    result = agent_executor.invoke({"input": input_text})
    output = result["output"]
    if isinstance(output, str):
        intermediate_results = InitialParseResponseV2.model_validate_json(output)
    elif isinstance(output, dict):
        intermediate_results = InitialParseResponseV2.model_validate(output)
    else:
        raise ValueError("Unable to parse LLM response")
    selection = [input_environment.concepts[x] for x in intermediate_results.selection]
    order = parse_order(selection, intermediate_results.order or [])

    filtering = (
        parse_filtering(selection, intermediate_results.filtering)
        if intermediate_results.filtering
        else None
    )
    # from trilogy.core.models import unique
    # concepts = unique(concepts, 'address')
    if debug:
        print("Concepts found")
        for c in intermediate_results.columns:
            print(c)
        if intermediate_results.order:
            print("Ordering")
            for o in intermediate_results.order:
                print(o)
    query = SelectStatement(
        selection=selection,
        limit=safe_limit(intermediate_results.limit),
        order_by=order,
        where_clause=filtering,
    )
    return query


def build_query(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    log_info: bool = True,
) -> ProcessedQuery:
    query = parse_query(
        input_text,
        input_environment,
        debug=debug,
        llm=llm,
        log_info=log_info,
    )
    return process_query(statement=query, environment=input_environment)



from trilogy_nlp.prompts.prompt_executor import SelectionPromptCase
from trilogy_nlp.models import FinalParseResponse
from tests.utility import generate_test_case, evaluate_cases


def gen_select_test(words, filters: list[str] | None = None):
    def select_test(x: FinalParseResponse):
        return set(x.selection) == set(words)

    if filters:

        def filter_test(x: FinalParseResponse):
            return set([z.concept for z in x.filtering]) == set(filters)  # type: ignore

        return [select_test, filter_test]
    return [select_test]


def test_selection_prompt(engine):
    test1 = generate_test_case(
        SelectionPromptCase,
        tests=gen_select_test(["question.creation_date.year", "question.id.count"]),
        question="How many questions are asked per year?",
        concept_names=[
            "question.creation_date.year",
            "question.id",
            "question.id.count",
            "question.author",
        ],
        llm=engine,
    )

    test2 = generate_test_case(
        SelectionPromptCase,
        tests=gen_select_test(["question.author"], ["question.id"]),
        question="Author of question id 2?",
        concept_names=[
            "question.creation_date.year",
            "question.id",
            "question.id.count",
            "question.author",
        ],
        llm=engine,
    )
    evaluate_cases([test1, test2])



from .prompt_executor import (
    FilterRefinementCase,
    SelectionPromptCase,
    SemanticExtractionPromptCase,
    SemanticToTokensPromptCase,
    FilterRefinementErrorCase,
    run_prompt,
)

__all__ = [
    "run_prompt",
    "SelectionPromptCase",
    "SemanticExtractionPromptCase",
    "SemanticToTokensPromptCase",
    "FilterRefinementCase",
    "FilterRefinementErrorCase",
]



from pydantic import BaseModel, RootModel
from trilogy.core.enums import ComparisonOperator, Ordering
from trilogy.core.models import Concept
from typing import Union

### Intermediate Models


class FilterResult(BaseModel):
    """The result of the filter prompt"""

    concept: str
    values: list[str | int | float | bool]
    operator: ComparisonOperator


class FinalFilterResult(BaseModel):
    concept: Concept
    values: list[str | int | float | bool]
    operator: ComparisonOperator


class OrderResult(BaseModel):
    """The result of the order prompt"""

    concept: str
    order: Ordering


class FinalOrderResult(BaseModel):
    """The processed result of the order prompt"""

    concept: Concept
    order: Ordering


class InitialParseResponse(BaseModel):
    """The result of the initial parse"""

    metrics: list[str]
    dimensions: list[str]
    limit: int
    order: list[OrderResult]
    filtering: list[FilterResult]

    @property
    def selection(self) -> list[str]:
        filtering = [f.concept for f in self.filtering]
        order = [x.concept for x in self.order]
        return list(set(self.metrics + self.dimensions + filtering + order))


class FinalParseResponse(BaseModel):
    """The result of the initial parse"""

    selection: list[str]
    limit: int
    order: list[OrderResult]
    filtering: list[FilterResult]


class IntermediateParseResults(BaseModel):
    select: list[Concept]
    limit: int
    order: list[FinalOrderResult]
    filtering: list[FinalFilterResult]


### Parse Result Models


class SemanticTokenMatch(BaseModel):
    phrase: str
    tokens: list[str]


class SemanticTokenResponse(RootModel):
    root: list[SemanticTokenMatch]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, idx):
        return self.root.__getitem__(idx)


class ConceptSelectionResponse(BaseModel):
    matches: list[str]
    reasoning: str


class FilterRefinementResponse(BaseModel):
    new_values: list[Union[str, float, int, bool]]
    old_values: list[str]
    reasoning: str



from trilogy_nlp.prompts.prompt_executor import SemanticToTokensPromptCase
from pydantic import ValidationError
import pytest


def test_retry(engine):
    test = SemanticToTokensPromptCase(
        tokens=["alphabet", "soup", "apple", "fruit", "shape", "orange"],
        phrases=["round fruits"],
        purpose="Dimension",
        llm=engine,
    )

    test.execute_prompt = (
        lambda x, skip_cache=False: test.prompt + "\nFORCED_SYNTAX_ERROR"
    )

    with pytest.raises(ValidationError):
        test._run(dry_run=False)

    assert test.retry_prompt in test.prompt



from trilogy_nlp.prompts import SemanticToTokensPromptCase
from trilogy_nlp.models import SemanticTokenMatch
from tests.utility import generate_test_case, evaluate_cases


def gen_semantic_tests(match_dict):
    output = []
    for key, tokens in match_dict.items():

        def test_case(input: list[SemanticTokenMatch]):
            candidate = [x for x in input if x.phrase == key]
            if not candidate:
                return False
            matched: SemanticTokenMatch = candidate[0]
            return set(tokens) == set(matched.tokens)

        output.append(test_case)
    return output


def test_structured_input(engine):
    first_phrase = "number of questions per year"
    test1 = generate_test_case(
        SemanticToTokensPromptCase,
        tests=gen_semantic_tests({first_phrase: ["question", "year", "count"]}),
        phrases=[first_phrase],
        tokens=["question", "year", "count", "albatross", "answer", "month", "quarter"],
        purpose="calculated metrics",
        llm=engine,
    )

    evaluate_cases([test1])



from trilogy import Environment
from trilogy_nlp.enums import Provider
from trilogy_nlp.main import build_query
from trilogy_nlp.main_v2 import build_query as build_query_v2


class NLPEngine(object):
    def __init__(
        self,
        provider: Provider,
        model: str | None = None,
        api_key: str | None = None,
        args: dict = {},
    ):
        self.provider = provider
        self.debug = False
        self.model = model
        self.api_key = api_key
        self.llm = self.create_llm()

    def create_llm(self):
        if self.provider == Provider.OPENAI:
            from langchain_community.chat_models import ChatOpenAI

            llm = ChatOpenAI(
                model_name=self.model if self.model else "gpt-3.5-turbo",
                openai_api_key=self.api_key,
                # model_kwargs=chat_openai_model_kwargs,
            )
        elif self.provider == Provider.GOOGLE:
            from langchain_google_genai import ChatGoogleGenerativeAI

            llm = ChatGoogleGenerativeAI(
                model=self.model if self.model else "gemini-pro",
                convert_system_message_to_human=True,
                google_api_key=self.api_key,
            )
        else:
            raise NotImplementedError(f"Unsupported provider {self.provider}")
        return llm

    def test_connection(self):
        local = self.create_llm()
        result = local.invoke("Hello")
        print(result.content)

    def build_query_from_text(
        self, input_text: str, env: Environment, version: int = 1
    ):
        if version == 2:
            return build_query_v2(
                input_text=input_text,
                input_environment=env,
                debug=self.debug,
                llm=self.llm,
            )
        return build_query(
            input_text=input_text,
            input_environment=env,
            debug=self.debug,
            llm=self.llm,
        )



from trilogy_nlp.models import SemanticTokenResponse, InitialParseResponse


def test_phrase_token():
    test = """[
   {
      "phrase":"questions per year",
      "tokens":[
         "question",
         "year"
      ]
   }
]"""

    x = SemanticTokenResponse.model_validate_json(test)

    assert x[0].tokens == ["question", "year"]


def test_initial_parse_result():
    response = """{
"metrics":["number of questions"],
"dimensions": ["year"],
"limit": -1,
"order": [],
"filtering": [{"concept":"year", "values":["2020"], "operator":"="}]
}"""

    x = InitialParseResponse.model_validate_json(response)
    assert x.limit == -1
